
/*----------------------------------------------------------------------------+
 |                                                                            |
 |                  Oregon Sensor Support for HEYU                            |
 |                Copyright 2008 Charles W. Sullivan                          |
 |                      All Rights Reserved                                   |
 |                                                                            |
 |                                                                            |
 | This software is licensed free of charge for non-commercial distribution   |
 | and for personal and internal business use only.  Inclusion of this        |
 | software or any part thereof in a commercial product is prohibited         |
 | without the prior written permission of the author.  You may copy, use,    |
 | and distribute this software subject to the following restrictions:        |
 |                                                                            |
 |  1)	You may not charge money for it.                                      |
 |  2)	You may not remove or alter this license, copyright notice, or the    |
 |      included disclaimers.                                                 |
 |  3)	You may not claim you wrote it.                                       |
 |  4)	If you make improvements (or other changes), you are requested        |
 |	to send them to the Heyu maintainer so there's a focal point for      |
 |      distributing improved versions.                                       |
 |                                                                            |
 | As used herein, HEYU is a trademark of Daniel B. Suthers.                  | 
 | X10, CM11A, and ActiveHome are trademarks of X-10 (USA) Inc.               |
 | The author is not affiliated with either entity.                           |
 |                                                                            |
 | Charles W. Sullivan                                                        |
 | Co-author and Maintainer                                                   |
 | Greensboro, North Carolina                                                 |
 | Email ID: cwsulliv01                                                       |
 | Email domain: -at- heyu -dot- org                                          |
 |                                                                            |
 | Disclaimers:                                                               |
 | THERE IS NO ASSURANCE THAT THIS SOFTWARE IS FREE OF DEFECTS AND IT MUST    |
 | NOT BE USED IN ANY SITUATION WHERE THERE IS ANY CHANCE THAT ITS            |
 | PERFORMANCE OR FAILURE TO PERFORM AS EXPECTED COULD RESULT IN LOSS OF      |
 | LIFE, INJURY TO PERSONS OR PROPERTY, FINANCIAL LOSS, OR LEGAL LIABILITY.   |
 |                                                                            |
 | TO THE EXTENT ALLOWED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED "AS IS",|
 | WITH NO EXPRESS OR IMPLIED WARRANTY, INCLUDING, BUT NOT LIMITED TO, THE    |
 | IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.|
 |                                                                            |
 | IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL THE AUTHOR BE LIABLE    |
 | FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL   |
 | DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SOFTWARE EVEN IF   |
 | THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.            |
 |                                                                            |
 +----------------------------------------------------------------------------*/



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <ctype.h>
#if defined(SYSV) || defined(FREEBSD) || defined(OPENBSD)
#include <string.h>
#else
#include <strings.h>
#endif

#include "x10.h"
#include "process.h"
#include "x10state.h"
#include "oregon.h"

extern int sptty;
extern int i_am_state, i_am_aux, i_am_relay, i_am_rfxmeter;
extern char *funclabel[];
extern void create_flagslist ( unsigned char, unsigned int, char * );
extern double celsius2temp ( double, char, double );

extern CONFIG config;
extern CONFIG *configp;

extern struct x10global_st x10global;
extern x10hcode_t *x10state;

extern unsigned int modmask[NumModMasks][16];
extern int lock_for_write(), munlock();

extern unsigned short ore_ignore[];
extern int            ore_ignore_size;

#if 0
int str2hex ( char *str, unsigned char *buf, int *bufsiz )
{
   char minibuf[16];
   char *sp;
   int  j;
   
   if ( (strlen(str) % 2) != 0 ) {
      fprintf(stderr, "Invalid string length %d\n", strlen(str));
      return 1;
   }

   sp = str;

   j = 0;
   while ( *sp != '\0' ) {
      minibuf[0] = *sp++;
      minibuf[1] = *sp++;
      minibuf[2] = '\0';
      buf[j++] = (unsigned char)strtol(minibuf, NULL, 16);
   }
   if ( buf[0] != 8 * (j - 1) ) {
      fprintf(stderr, "Miscount: buf[0] = %d, bufsiz = %d\n", buf[0]/8, j - 1);
      return 1;
   }

   memmove(buf, buf + 1, j - 1);
   *bufsiz = j - 1;

   return 0;
}

unsigned int bcd2dec ( unsigned int bcd, int digits )
{
   unsigned int dec = 0, mult = 1, mask =0x0f;
   int          j;

   for ( j = 0; j < digits; j++ ) {
      dec += mult * ((bcd & mask) >> (4 * j));
      mult *= 10;
      mask <<= 4;
   }
   return dec;
}
#endif


int is_ore_ignored ( unsigned short saddr )
{
#ifdef HASORE
   int j = 0;

   while ( j < ore_ignore_size ) {
      if ( saddr == ore_ignore[j] )
         return 1;
      j++;
   }
#endif
   return 0;
}

#ifdef HASORE
unsigned char nybble_sum( int nbytes, unsigned char *buf )
{
   int   j;
   unsigned char sum = 0;

   for ( j = 0; j < nbytes; j++ ) {
      sum += (buf[j] & 0x0f) + ((buf[j] >> 4) & 0x0f);
   }
   return sum;
}
      
unsigned char cse ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(7, buf) - buf[7];
   
   return (unsigned char)(ssum & 0xff) - 0x18u;
}   

unsigned char cs7 ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(7, buf) - buf[7];
   
   return (unsigned char)(ssum & 0xff) - 0x0Au;
}   

unsigned char cs8 ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(8, buf) - buf[8];
   
   return (unsigned char)(ssum & 0xff) - 0x0Au;
}

unsigned char cs9 ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(9, buf) - buf[9];
   
   return (unsigned char)(ssum & 0xff) - 0x0Au;
}   

unsigned char cs10 ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(10, buf) - buf[10] ;
   
   return (unsigned char)(ssum & 0xff) - 0x0Au;
}

unsigned char cs11 ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(11, buf) - buf[11] ;
   
   return (unsigned char)(ssum & 0xff) - 0x0Au;
}

unsigned char cs2HL ( unsigned char *buf )
{
  short ssum;

  ssum = nybble_sum(8, buf) + (buf[8] & 0x0f) -
        ((buf[8] >> 4) & 0x0f) - ((buf[9] << 4) & 0xf0);

  return (unsigned char)(ssum & 0xff) - 0x0Au;
}

unsigned char csRHL ( unsigned char *buf )
{
  short ssum;

  ssum = nybble_sum(9, buf) + (buf[9] & 0x0f)
         - ((buf[9] >> 4) & 0x0f) - ((buf[10] << 4) & 0xf0);
  
  return (unsigned char)(ssum & 0xff) - 0x0A;
}

unsigned char csWHL ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(6, buf) + (buf[6] & 0x0f) - 
          ((buf[6] >> 4) & 0x0f) - ((buf[7] << 4) & 0xf0);

   return (unsigned char)(ssum & 0xff) - 0x0Au;
}

unsigned char csNull ( unsigned char *buf )
{
  return 0;
}

/* Types
GR101:    GR101
TEMP1:    THR128,THx138
TEMP2:    THN132N,THWR288
TH1:      THGN122N,THGR122NX,THGR228N,THGR268
TH2:      THGR810
TH3:      RTGR328N
TH4: THGR328 (WIP)
TH5:      WTGR800
TH6:      THGR918
THB1:     BTHR918
THB2:     BTHR918N,BTHR968
RAIN1:    RGR126,RGR682,RGR918
RAIN2:    PCR800
RAIN3:
WIND1:    WTGR800
WIND2:    WGR800
WIND3:    STR918,WGR918
UV1:      UVR138
UV2:      UVN800
DT1:      RTGR328N
ELEC1:    cent-a-meter
WEIGHT1:  BWR101,BWR102
*/


static struct oregon_st {
   unsigned char  minbits;
   unsigned char  trulen;
   unsigned char  id0mask;
   unsigned char  id1mask;
   unsigned char  id0;
   unsigned char  id1;
   int            nvar;
   unsigned char  chsize;
   unsigned char  batmode;
   unsigned short bpoffset;
   unsigned char  is_supported;
   unsigned char  addr_mask;
   int            addr_byte;
   char           *slabel;
   unsigned char  subtype;
   unsigned char  (*chksumf)(unsigned char *);
} orechk[] = {
   {72,   9, 0xff, 0xff, 0x0a, 0x4d, 1, 3, 1, 0,   1, 0xff, 4, "T1",   OreTemp1,   cs8   },
   {60,   8, 0xff, 0xff, 0xea, 0x4c, 1, 3, 1, 0,   1, 0xff, 4, "T2",   OreTemp2,   csWHL },
   {72,   9, 0xff, 0xff, 0x1a, 0x2d, 2, 3, 1, 0,   1, 0xff, 4, "TH1",  OreTH1,     cs8   },
   {72,   9, 0xff, 0xff, 0xfa, 0x28, 2, 4, 1, 0,   1, 0xff, 4, "TH2",  OreTH2,     cs8   },
   {72,   9, 0x0f, 0xff, 0x0a, 0xcc, 2, 4, 1, 0,   1, 0xff, 4, "TH3",  OreTH3,     cs8   },
   {72,   9, 0x0f, 0xff, 0x0a, 0xcc, 2, 3, 1, 0,   1, 0xff, 4, "TH4",  OreTH4,     cs8   },
   {72,   9, 0xff, 0xff, 0xfa, 0xb8, 2, 0, 2, 0,   1, 0xff, 4, "TH5",  OreTH5,     cs8   },
   {72,   9, 0xff, 0xff, 0x1a, 0x3d, 2, 3, 2, 0,   1, 0xff, 4, "TH6",  OreTH6,     cs8   },
   {88,  11, 0xff, 0xff, 0x5a, 0x5d, 3, 3, 1, 795, 1, 0xff, 4, "THB1", OreTHB1,    cs10  },
   {88,  11, 0xff, 0xff, 0x5a, 0x6d, 3, 3, 2, 856, 1, 0xff, 4, "THB2", OreTHB2,    cs10  },


   {80,  10, 0xff, 0xff, 0x2a, 0x1d, 2, 0, 2, 0,   0, 0xff, 4, "R1",   OreRain1,   cs2HL },
   {84,  11, 0xff, 0xff, 0x2a, 0x19, 2, 0, 2, 0,   0, 0xff, 4, "R2",   OreRain2,   csRHL },
   {84,  11, 0xff, 0xff, 0x06, 0xe4, 2, 3, 1, 0,   0, 0xff, 4, "R3",   OreRain3,   csRHL },
   {80,  10, 0xff, 0xff, 0x1a, 0x99, 2, 0, 2, 0,   0, 0xff, 4, "WS1",  OreWind1,   cs9   },
   {80,  10, 0xff, 0xff, 0x1a, 0x89, 2, 0, 2, 0,   0, 0xff, 4, "WS2",  OreWind2,   cs9   },
   {80,  10, 0xff, 0xff, 0x3a, 0x0d, 3, 0, 2, 0,   0, 0xff, 4, "WS3",  OreWind3,   cs9   },
   {60,   8, 0xff, 0xff, 0xea, 0x7c, 1, 0, 1, 0,   0, 0xff, 4, "UV1",  OreUV1,     csWHL },
   {64,   8, 0xff, 0xff, 0xda, 0x78, 1, 0, 1, 0,   0, 0xff, 4, "UV2",  OreUV2,     cs7   },
   {96,  12, 0xff, 0xff, 0x8a, 0xec, 2, 4, 2, 0,   0, 0xff, 4, "DT1",  OreDT1,     cs11  },
   {64,   8, 0xff, 0xf0, 0xea, 0x00, 3, 0, 0, 0,   0, 0xff, 4, "EL1",  OreElec1,   cse   },
   {64,   8, 0xff, 0xf0, 0xea, 0x20, 3, 0, 0, 0,   0, 0xff, 4, "EL1",  OreElec1,   cse   },
   {56,   7, 0x00, 0x00, 0x00, 0x00, 2, 0, 0, 0,   0, 0xf0, 2, "WG1",  OreWeight1, csNull},
   {64,   8, 0x0f, 0x00, 0x03, 0x00, 2, 0, 0, 0,   0, 0xf0, 2, "WG2",  OreWeight2, csNull},    /* OreGR101 */

};
int orechk_size = sizeof(orechk)/sizeof(struct oregon_st);


unsigned char channelval( char *chstr, unsigned char chsize,
                                       unsigned char channel )
{
   unsigned char chval;

   switch ( chsize ) {
      case 3 :
         channel &= 0x07;
         switch ( channel ) {
            case 1 :  chval = 1; sprintf(chstr, "Ch 1 "); break;
            case 2 :  chval = 2; sprintf(chstr, "Ch 2 "); break;
            case 4 :  chval = 3; sprintf(chstr, "Ch 3 "); break;
            default : chval = 0; sprintf(chstr, "Ch ? "); break;
         }
         break;
      case 4 :
         chval = channel;
         sprintf(chstr, "Ch %d ", channel);
         break;
      default :
         chval = 0;
         *chstr = '\0';
         break;
   }
   return chval;
}
#endif /* HASORE */

 
unsigned char oretype ( unsigned char *xbuf, unsigned char *subindx,
                        unsigned char *trulen, unsigned short *addr, int *nseq )
{
#ifdef HASORE
   int j = 0;

   while ( j < orechk_size ) {
      if ( (xbuf[0] >= orechk[j].minbits) &&
           (xbuf[1] & orechk[j].id0mask) == orechk[j].id0 &&
           (xbuf[2] & orechk[j].id1mask) == orechk[j].id1 &&
           orechk[j].chksumf(xbuf + 1) == 0 ) {
         *trulen = orechk[j].trulen;
         *addr = (unsigned short)(xbuf[orechk[j].addr_byte] & orechk[j].addr_mask);
         *nseq = (int)orechk[j].nvar;
         *subindx = j;
         if ( orechk[j].is_supported || configp->disp_ore_all == YES ) {
            return 0;
         }
         else {
            return 1;
         }
      }
      j++;
   }
   return 1;
#else
   return 1;
#endif /* HASORE */
}


/*----------------------------------------------------------------------------+
 | Translate Oregon RF data                                                   |
 | Buffer reference: ST_COMMAND, ST_GENLONG, plus:                            |
 |  vtype, subindx, vidhi, vidlo, seq, buflen, buffer                         |
 |   [2]     [3]     [4]    [5]   [6]    [7]    [8+]                          |
 +----------------------------------------------------------------------------*/
char *translate_oregon( unsigned char *buf, unsigned char *sunchanged, int *launchp )
{
#ifdef HASORE
   static char    outbuf[160];
   char           flagslist[80];
   ALIAS          *aliasp;
   LAUNCHER       *launcherp;
   extern unsigned int signal_source;
   unsigned char  func, *vdatap, vtype, subindx, subtype, seq, channel_code;
   int            humid, lasthumid;
   int            baro, lastbaro;
   unsigned short vident;
   int            unit, loc;
   int            dtempc, lastdtempc, delta, nvar;
   int            buflen;
   unsigned long  longvdata;

   char           hc;

   unsigned char  actfunc, xactfunc;
   unsigned int   trigaddr, mask, active, trigactive;
   unsigned int   bmaplaunch;
   unsigned long  afuncmap, xfuncmap;

   int            j, k, found, index = -1;
   unsigned char  hcode, ucode, trig = 0;
   unsigned int   bitmap, vflags = 0;

   static unsigned long prevdata[3], lastdata[3];
   static char          chstr[8];
   static unsigned char channel;
   static unsigned char blevel;
   static unsigned char status;
   static char          batstr[32];

   double temp, dbaro;

   static char *typename[] = {"Std", "Ent", "Sec", "RFXSensor", "RFXMeter",
     "?", "??", "???", "????", "DigiMax", "Oregon", "Noise", "Noise2"};

   launcherp = configp->launcherp;
   aliasp = configp->aliasp;

   *launchp = -1;
   *sunchanged = 0;
   flagslist[0] = '\0';
   *outbuf = '\0';


   vtype   = buf[2];
   subindx = buf[3];
   vident  = (unsigned short)buf[5];
   seq     = buf[6];
   buflen  = buf[7];
   vdatap  = buf + 8;
   func    = VdataFunc;

   subtype = orechk[subindx].subtype;

   x10global.longvdata = 0;
   x10global.lastvtype = vtype;

   hcode = ucode = 0;
   found = 0;
   j = 0;
   /* Look up the alias, if any */
   while ( !found && aliasp && aliasp[j].line_no > 0 ) {
      if ( aliasp[j].vtype == vtype && aliasp[j].subtype == subtype && 
           (bitmap = aliasp[j].unitbmap) > 0 &&
           (ucode = single_bmap_unit(bitmap)) != 0xff ) {
         for ( k = 0; k < aliasp[j].nident; k++ ) {
            if ( aliasp[j].ident[k] == vident ) {
               index = j;
               found = 1;
               break;
            }
         }
      }
      j++;
   }

   if ( !found || !aliasp ) {
      if ( seq == 1 ) {
         sprintf(outbuf, "func %11s : Type %s_%s ID 0x%02X Data 0x",
            "RFdata", typename[vtype], orechk[subindx].slabel, vident);
         for ( j = 0; j < buflen; j++ ) {
            sprintf(outbuf + strlen(outbuf), "%02x", vdatap[j]);
         }
      }
      return outbuf;
   }

   nvar = orechk[subindx].nvar;

   /* Internal check */
   if ( aliasp[index].storage_units != (2 * nvar) ) {
      sprintf(outbuf, "Internal error - storage != 2 * nvar\n");
      return outbuf;
   }
      
   
   hc = aliasp[index].housecode;
   hcode = hc2code(hc);
   unit = code2unit(ucode);

   /* Update activity timeout for sensor */
   aliasp[index].timeout = configp->inactive_timeout;

   switch ( subtype ) {
      case OreTH1 :
      case OreTH2 :
      case OreTH3 :
      case OreTH4 :
      case OreTH5 :
      case OreTH6 :
      case OreTemp1 :
      case OreTemp2 :
      case OreTHB1 :
      case OreTHB2 :
         loc = aliasp[index].storage_index;
         if ( seq == 1 ) {
            /* Process temperature and battery data */
            for ( j = 0; j < nvar; j++ ) {
               prevdata[j] = x10global.data_storage[loc + j];
               lastdata[j] = x10global.data_storage[loc + nvar + j];
            }
            channel_code = (vdatap[2] >> 4) & 0x0f;
            channel = channelval(chstr, orechk[subindx].chsize, channel_code);
            longvdata = (unsigned long)(channel & 0x0f) << 28;

            dtempc = 100 * ((vdatap[5] & 0xf0) >> 4) +
                            10 * (vdatap[5] & 0x0fu) +
                                ((vdatap[4] & 0xf0u) >> 4);
            longvdata |= (unsigned long)dtempc << 8;
            status = 0x01;

            if ( vdatap[6] & 0x08u ) {
               dtempc = -dtempc;
               status |= ORE_NEGTEMP;
            }

            switch ( orechk[subindx].batmode ) {
               case 2 :
                  blevel = 10 - (vdatap[4] & 0x0fu);
                  sprintf(batstr, "BatLvl %d%% ", blevel * 10);
                  status |= ORE_BATLVL;
                  if ( (10 * blevel) <= configp->ore_lobat ) {
                     status |= ORE_LOBAT;
                     sprintf(batstr + strlen(batstr), "LoBat ");
                  }
                  break;
               case 1 :
                  blevel = 0;               
                  if ( vdatap[4] & 0x04u ) {
                     status |= ORE_LOBAT;
                     strcpy(batstr, "LoBat ");
                  }
                  else {
                     blevel = 0;
                     *batstr = '\0';
                  }
                  break;
               default :
                  blevel = 0;
                  *batstr = '\0';
                  break;
            }
                     
            longvdata |= status | (blevel << 24);

            lastdtempc = (lastdata[0] & 0x000fff00u) >> 8;
            if ( lastdata[0] & ORE_NEGTEMP ) 
               lastdtempc = -lastdtempc;
            delta = abs(dtempc - lastdtempc);
            *sunchanged = (delta < (int)configp->ore_chgbits_t) ? 1 : 0;

            func = OreTempFunc;
            trig = OreTempTrig;
            vflags |= (status & ORE_LOBAT) ? SEC_LOBAT : 0;

            temp = celsius2temp((double)dtempc / 10.0, configp->ore_tscale, configp->ore_toffset);
            sprintf(outbuf, "func %11s : hc %c unit %d %sTemp "FMT_ORET"%c %s(%s)",  
               funclabel[func], hc, unit, chstr, temp, configp->ore_tscale,
               batstr /*((vflags & SEC_LOBAT) ? "LoBat " : "")*/, aliasp[index].label);
         }
         else if ( seq == 2 ) {
            /* Process Humidity data */
            longvdata = (unsigned long)(channel & 0x0f) << 28;
            humid = 10 * (vdatap[7] & 0x0f) + ((vdatap[6] & 0xf0) >> 4);
            longvdata |= (unsigned char)humid << 8;
            status |= 0x01;

            longvdata |= status | (blevel << 24);

            lasthumid = (lastdata[seq - 1] & 0x0000ff00u) >> 8;             
            delta = abs(humid - lasthumid);
            *sunchanged = (delta < (int)configp->ore_chgbits_rh) ? 1 : 0;

            func = OreHumidFunc;
            trig = OreHumidTrig;
            vflags |= (status & ORE_LOBAT) ? SEC_LOBAT : 0;

            sprintf(outbuf, "func %11s : hc %c unit %d %sRH %d%% %s(%s)",
               funclabel[func], hc, unit, chstr, humid,
               batstr /*((vflags & SEC_LOBAT) ? "LoBat " : "")*/, aliasp[index].label);
         }
         else if ( seq == 3 ) {
            /* Process Barometric Pressure data */
            longvdata = (unsigned long)(channel & 0x0f) << 28;
            baro = (int)vdatap[8] + orechk[subindx].bpoffset;
            longvdata |= (unsigned char)baro << 8;
            status |= 0x01;

            longvdata |= status | (blevel << 24);

            lastbaro = (lastdata[seq - 1] & 0x0000ff00u) >> 8;
            delta = abs(baro - lastbaro);
            *sunchanged = (delta < (int)configp->ore_chgbits_bp) ? 1 : 0;

            func = OreBaroFunc;
            trig = OreBaroTrig;
            vflags |= (status & ORE_LOBAT) ? SEC_LOBAT : 0;

            dbaro = (double)baro * configp->ore_bpscale + configp->ore_bpoffset;
            sprintf(outbuf, "func %11s : hc %c unit %d %sBP "FMT_OREBP"%s %s(%s)",
               funclabel[func], hc, unit, chstr, dbaro, configp->ore_bpunits,
               batstr /*((vflags & SEC_LOBAT) ? "LoBat " : "")*/, aliasp[index].label);            
         }
         else {
            break;
         }

         x10global.longvdata = longvdata;

         x10global.data_storage[loc + seq - 1] = longvdata;
         x10state[hcode].longdata = longvdata;

         if ( *sunchanged == 0 ) {
            x10state[hcode].state[ChgState] |= bitmap;
            x10global.data_storage[loc + nvar + seq - 1] = longvdata;
         }
         else {
            x10state[hcode].state[ChgState] &= ~bitmap;
         }

#if 0
         sprintf(outbuf + strlen(outbuf), " %s", ((*sunchanged) ? "NOCHG":"CHG"));
#endif

         break;


      default :
         break;
   }

   x10state[hcode].vaddress = bitmap;
   x10state[hcode].lastcmd = func;
   x10state[hcode].lastunit = unit;
   x10state[hcode].vident[ucode] = vident;
   x10state[hcode].vflags[ucode] = vflags;
   x10state[hcode].timestamp[ucode] = time(NULL);
   x10global.lasthc = hcode;
   x10global.lastaddr = 0;


   actfunc = 0;
   afuncmap = 0;
   xactfunc = func;
   xfuncmap = (1 << trig);
   trigaddr = bitmap;

   mask = modmask[VdataMask][hcode];
   active = bitmap & mask;

   if ( i_am_state )
      write_x10state_file();

   /* Heyuhelper, if applicable */
   if ( i_am_state && signal_source & RCVI && configp->script_mode & HEYU_HELPER ) {
      launch_heyuhelper(hcode, trigaddr, func);
      return 0;
   }

   bmaplaunch = 0;
   j = 0;
   while ( launcherp && launcherp[j].line_no > 0 ) {
      if ( (launcherp[j].flags & x10global.flags) != launcherp[j].flags ||
	   (launcherp[j].notflags & ~x10global.flags) != launcherp[j].notflags ||
           (launcherp[j].vflags & x10state[hcode].vflags[ucode]) != launcherp[j].vflags ||
	   (launcherp[j].notvflags & ~x10state[hcode].vflags[ucode]) != launcherp[j].notvflags ||
           (launcherp[j].czflags & x10global.czflags) != launcherp[j].czflags ||
	   (launcherp[j].notczflags & ~x10global.czflags) != launcherp[j].notczflags ||
            launcherp[j].type != L_NORMAL ) {
         j++;
	 continue;
      }

      if ( launcherp[j].hcode == hcode &&
           launcherp[j].xfuncmap & xfuncmap &&
           launcherp[j].source & signal_source ) {
         trigactive = trigaddr & (mask | launcherp[j].signal);
         if ( (bmaplaunch = (trigactive & launcherp[j].bmaptrig) & 
                   (x10state[hcode].state[ChgState] | ~launcherp[j].chgtrig)) ||
                   (launcherp[j].unitstar && !trigaddr)) {
            *launchp = j;
            launcherp[j].actfunc = xactfunc;
            launcherp[j].genfunc = 0;
	    launcherp[j].xfunc = xactfunc;
	    launcherp[j].level = x10state[hcode].dimlevel[ucode];
            launcherp[j].bmaplaunch = bmaplaunch;
            launcherp[j].actsource = signal_source;
            break;
         }
      }
      j++;
   }

   x10state[hcode].launched = bmaplaunch;

   return outbuf;
#else
   return "";
#endif /* HASORE */
}   

#ifdef HASORE
/*---------------------------------------------------------------------+
 | Display stored data for Oregon sensors                              |
 +---------------------------------------------------------------------*/
int show_oregon ( void )
{

   ALIAS          *aliasp;
   char           hc;
   int            j, unit, index, count = 0, maxlabel = 0, maxmod = 0;
   int            loc, locunits;
   unsigned char  hcode, ucode, func;
   unsigned char  subtype, subindx, status = 0;
   unsigned long  longvdata = 0;
   short          dtempc;
   int            baro, blevel;
   double         temp, dbaro;

   if ( !(aliasp = configp->aliasp) ) 
      return 0;

   /* Get maximum lengths of module name and alias label */
   index = 0;
   while ( aliasp[index].line_no > 0 ) {
      if ( aliasp[index].vtype == RF_OREGON ) {
         count++;
         maxlabel = max(maxlabel, (int)strlen(aliasp[index].label));
         maxmod = max(maxmod, (int)strlen(lookup_module_name(aliasp[index].modtype)) );
      }
      index++;
   }

   if ( count == 0 )
      return 0;

   index = 0;
   while ( aliasp[index].line_no > 0 ) {
      if ( aliasp[index].vtype != RF_OREGON ) {
         index++;
         continue;
      }
      ucode = single_bmap_unit(aliasp[index].unitbmap);
      unit = code2unit(ucode);
      hc = aliasp[index].housecode;
      hcode = hc2code(hc);
      subtype = aliasp[index].subtype;
      loc = aliasp[index].storage_index;
      locunits = aliasp[index].storage_units;

      for ( subindx = 0; subindx < orechk_size; subindx++ ) {
         if ( orechk[subindx].subtype == subtype )
            break;
      }

      printf("%c%-2d %-*s %-*s ", hc, unit,
          maxmod, lookup_module_name(aliasp[index].modtype),
          maxlabel, aliasp[index].label); 

      for ( j = 0; j < locunits; j++ ) {
         func = aliasp[index].funclist[j];
         longvdata = x10global.data_storage[loc + j];
         status = longvdata & 0xff;
         switch ( func ) {
            case OreTempFunc :
               if ( orechk[subindx].chsize > 0 ) {
                  if ( !longvdata ) {
                     printf("Ch - ");
                  }
                  else {
                     printf("Ch %d ", (int)((longvdata & 0xf0000000) >> 28));
                  }
               } 

               if ( !longvdata ) {
                  printf("Temp ----%c ", configp->ore_tscale);
                  break;
               }

               dtempc = (longvdata & 0x000fff00) >> 8;
               if ( status & ORE_NEGTEMP )
                  dtempc = -dtempc;
               temp = celsius2temp((double)dtempc / 10.0, configp->ore_tscale, configp->ore_toffset);
               printf("Temp "FMT_ORET"%c ", temp, configp->ore_tscale);
               break;
            case OreHumidFunc :
               if ( !longvdata ) {
                  printf("RH --%% ");
                  break;
               }
               printf("RH %2d%% ", (int)(longvdata & 0xff00) >> 8);
               break;
            case OreBaroFunc :
               if ( !longvdata ) {
                  printf("BP ---%s", configp->ore_bpunits);
                  break;
               }
               baro = (int)(((longvdata & 0xff00) >> 8) + orechk[subindx].bpoffset);
               dbaro = (double)baro * configp->ore_bpscale + configp->ore_bpoffset;
               printf("BP "FMT_OREBP"%s ", dbaro, configp->ore_bpunits);
               break;
         }
      }

      if ( status & ORE_BATLVL ) {
         blevel = 10 * (int)((longvdata & 0x0f000000) >> 24);
         printf("BatLvl %d%% ", blevel );
         if ( blevel <= configp->ore_lobat )
            printf("LoBat ");
      }
      else if ( status & ORE_LOBAT ) {
         printf("LoBat ");
      }

      printf("\n");
      index++;
   }

   return 0;
}
#endif /* HASORE */

#ifdef HASORE 
/*---------------------------------------------------------------------+
 | Display a Digimax data value stored in the x10state structure.      |
 +---------------------------------------------------------------------*/
int c_orecmds ( int argc, char *argv[] )
{

   ALIAS          *aliasp;
   unsigned char  hcode, ucode, func;
   unsigned long  aflags;
   unsigned long  longvdata;
   char           hc;
   unsigned int   bitmap;
   int            j;
   int            unit, index, dtempc, baro, loc;
   double         temp, dbaro;

   static struct {
      char          *cmd;
      unsigned char func;
   } cmdlist[] = {
      {"oretemp",   OreTempFunc  },
      {"orerh",     OreHumidFunc },
      {"orebp",     OreBaroFunc  },
      {"",          0xff         },
   };

   int read_x10state_file ( void );

   if ( argc < 3 ) {
      fprintf(stderr, "Usage: %s %s Hu\n", argv[0], argv[1]);
      return 1;
   }

   if ( check_for_engine() != 0 ) {
      fprintf(stderr, "State engine is not running.\n");
      return 1;
   }
   if ( read_x10state_file() != 0 ) {
      fprintf(stderr, "Unable to read state file.\n");
      return 1;
   }

   if ( (aliasp = configp->aliasp) == NULL )
      return 1;

   j = 0;
   func = 0xff;
   while ( *cmdlist[j].cmd ) {
      if ( strcmp(argv[1], cmdlist[j].cmd) == 0 ) {
         func = cmdlist[j].func;
         break;
      }
      j++;
   }
      
   aflags = parse_addr(argv[2], &hc, &bitmap);
   if ( !(aflags & A_VALID) || aflags & (A_DUMMY | A_PLUS | A_MINUS) || bitmap == 0 ) {
      fprintf(stderr, "Invalid Hu address '%s'\n", argv[2]);
      return 1;
   }
   if ( aflags & A_MULT ) {
      fprintf(stderr, "Only a single unit address is valid.\n");
      return 1;
   }
   hcode = hc2code(hc);
   ucode = single_bmap_unit(bitmap);
   unit  = code2unit(ucode);

   if ( (index = alias_lookup_index(hc, bitmap, RF_OREGON)) < 0 ) {
      fprintf(stderr, "Address %c%d is not configured as an Oregon sensor\n", hc, unit);
      return 1;
   }

   loc = -1;
   for ( j = 0; j < aliasp[index].storage_units; j++ ) {
      if ( aliasp[index].funclist[j] == func ) {
         loc = aliasp[index].storage_index + j;
         break;
      }
   }
   if ( loc < 0 ) {
      fprintf(stderr, "Command '%s' not valid for this sensor.\n", argv[1]);
      return 1;
   }

   longvdata = x10global.data_storage[loc];

   if ( !longvdata ) {
      fprintf(stderr, "Not ready\n");
      return 1;
   }

   switch ( func ) {
      case OreTempFunc :
         dtempc = (longvdata & 0x000fff00u) >> 8;
         if ( longvdata & ORE_NEGTEMP )
            dtempc = -dtempc;
         temp = celsius2temp((double)dtempc / 10.0, configp->ore_tscale, configp->ore_toffset);
         printf("%.1f\n", temp);
         break;
     case OreHumidFunc :
         printf("%d\n", (int)((longvdata & 0x0000ff00u) >> 8));
         break;
     case OreBaroFunc :
         baro = (longvdata & 0x0000ff00u) >> 8;
         dbaro = (double)baro * configp->ore_bpscale + configp->ore_bpoffset;
         printf("%.1f\n", dbaro);
         break;
     default :
         break;
  }
  return 0;
}

#endif /* HASORE */
                          

